title: RAE2822 airfoil

# Transonic turbulent flow of Mach 0.72, Reynolds number 6.5 million, and angle of attack 2.92 deg over the RAE2822

<figure>
  <div style="display: flex; flex-wrap: wrap; justify-content: center; align-items:center">
    <img src="|media|/tutorials/RAE2822Pressure.png" alt="Contours" style="width:350px">
  </div>
  <div style="display: flex; flex-wrap: wrap; justify-content: center; align-items:center">
    <figcaption> Fig.1 Contour of pressure for subsonic flow over RAE2822 airfoil.</figcaption>
  </div>
</figure>


## Problem Statement
In this test case, a transonic turbulent flow of Mach 0.72 over the RAE2822 will be simulated. The domain used, and boundary
condition applied to the domain are illustrated in Fig. 2. The case definition and grid used were obtained from
[NPARC Alliance Validation Archive](https://www.grc.nasa.gov/WWW/wind/valid/raetaf/raetaf01/raetaf01.html). 
<figure>
  <div style="display: flex; flex-wrap: wrap; justify-content: center; align-items:center">
    <img src="|media|/tutorials/RAE2822Domain.png" alt="Contours" style="width:350px">
  </div>
  <div style="display: flex; flex-wrap: wrap; justify-content: center; align-items:center">
    <figcaption> Fig.2 Domain and boundary condition for RAE2822 airfoil test case.</figcaption>
  </div>
</figure>


## Mesh
A structured C-grid of size 369 × 65 x 2 (with 305 grid points on the airfoil)
 will be used, as shown in Fig. 3. The grid is available in the 
tutorial folder __|rootFolder|/run/Tutorial/RAE2822/CreateBlocks/__. 
@note
If `run` folder is empty, please download the content from [Github](https://github.com/FEST3D/run) 
direcotry or download the zip file [here](https://github.com/FEST3D/run/archive/master.zip).
@endnote
In the ```blocking_point.f90```
edit the number of blocks in the I-direction and J-direction. In this test case, four blocks will
be used. <br>
```integer, parameter :: xblocks = 6```<br>
```integer, parameter :: yblocks = 1```<br>
To compile the `blocking_point.f90` code with gfortran and execute it, commands are written 
in the makefile. Just use the following command to generate the grid files in the __grid/__ folder.<br>
```
$make
```
Run the previous command in the __|rootFolder|/run/Tutorial/RAE2822/CreateBlocks/__ directory.
<figure>
  <div style="display: flex; flex-wrap: wrap; justify-content: center; align-items:center">
    <img src="|media|/tutorials/RAE2822Grid.png" alt="mesh" style="width:350px">
  </div>
  <div style="display: flex; flex-wrap: wrap; justify-content: center; align-items:center">
    <figcaption> Fig.3 The Mesh for turbulent flow over flat plate test case.</figcaption>
  </div>
</figure>

## Setup
To setup the case directory, a python automation script is provided: ```automaton.py```. First, setup the most important
parameters, the paths to the grid files and main executable binary to FEST-3D<br>
```RunDir = 'RAE2822'```   Name of the run directory to create for current case. <br>
```GridDir= 'CreateBlocks/grid/'```  Path to the folder which only contains grid file. <br>
```NumberOfBlocks = 6``` Total number of blocks. It should match with number of gridfiles avaiable in the __GridDir__ folder <br>
```
AbsBinaryPath="/home/usr/FEST3D/bin/FEST3D"
``` 
Absolute path to the FEST-3D binary. Should be in __|rootfolder|/bin/FEST3D__<br>
Now provide the common control parameters.<br>
```Control['CFL'] = 100.0``` High CFL since implicit time-integration method will be used.<br>
```Control['LoadLevel'] = 0 ```Since simulation will be started from scratch, __0__ is specified.<br>
```Control['MaxIterations'] = 10000``` Maximum number of iteration to perform. <br>
```Control['SaveIterations'] = 1000``` Solution folder will be written every 1000 iteration. <br>
```Control['OutputFileFormat'] = 'vtk'``` Type of solution file to write. If you have tecplot file viewer you can user __tecplot__ instead of __vtk__<br>
```Control['Purge'] = 1``` Only one latest solution folder will be kept in the __time_directories__ and rest will be deleted. <br>
```Control['ResidualWriteInterval'] = 20 ```Write residual in __time_directories/aux/resnorm__ after every 5 iteration. <br>
```Control['Tolerance'] = "1e-13 Continuity_abs"``` Stop the iteration if the absolute residual value of _continuity_ equation is less than 1e-13. <br>
Few scheme parameters for inviscid flow:<br>
```Scheme['InviscidFlux'] = 'ausm'``` Inviscid flux-reconstruction shceme. You can use: __slau__, __ldfss0__, __ausmP__ , and __ausmUP instead of __ausm__.<br>
```Scheme['FaceState'] = 'muscl'``` Higher-order  face-state reconstruction method. you can use: __none__, __ppm__, and __weno__. <br>
```Scheme['Limiter'] = '1 1 1  0 0 0'``` Switch on the limiter for I,J,and K direction and switch off pressure based switching for all direction.<br>
```Scheme['TurbulenceLimiter'] = '1 1 1'``` Using limiter only for turbulent variables face-state reconstruction improve convergence.<br>
```Scheme['TurbulenceModel']='sst'``` using SST turbulence model. Other models __kkl__, __sa__, and __sst2003__ can be user also. <br>
```Scheme['TransitionModel']='none'```  __none__ is metioned for no transition model. <br>
```Scheme['TimeStep']='l'``` Local time stepping method. You can use global time-stepping method also __g__.<br>
```Scheme['TimeIntegration']='implicit'``` LU-SGS matrix-free time integration method.<br>
Now, lets define the flow feature of test case:<br>
```Flow["DensityInf"] = 1.2``` Free-stream density.<br>
```Flow["UInf"] = 252.9``` Free-stream x-component of velcotiy vector. (Mach = 0.5 at T_ref=300K)<br> 
```Flow["VInf"] = 10.2``` Free-stream y-component of velcotiy vector. <br>
```Flow["WInf"] = 0.0``` Free-stream z-component of velcotiy vector. <br>
```Flow["PressureInf"] = 103338.0``` Free-stream pressure. <br>
```Flow["TurbulenceIntensity"] = 1.0``` Free-stream turbulence intensity. <br>
```Flow["ViscosityRatio"] = 10.0``` Free-stream viscosity ratio. <br>
```Flow["Intermittency"] = 1.0``` Free-stream intermittency. <br>
```Flow["ReferenceViscosity"] = 1.4243-5``` Reference viscosity is set such that the Reynolds number (L_ref=1m) is 5 million<br>
    Flow["ViscosityLaw"] = "sutherland_law"``` Viscosity variation law. <br>

```OutputControl['Out'] = ["Velocity", "Density", "Pressure", "Mu"]``` Variables to write in the output file.<br>
```ResidualControl['Out'] = ["Mass_abs", "Viscous_abs", "Continuity_abs"]``` Residual to write in the resnorm file. <br>
```BoundaryConditions = [-4, -4, -5, -8, -6, -6]``` Broad boundary condition.[subsonic outlet, no-slip wall, far-field, and slip wall for rest ]. <br>

Rest of the variables should be left to their default value. To execute this script, use the following command:
```
$python automaton.py
```
Now, you will see a new folder created with ```RunDir``` name (_RAE2822_). Switch to that directory to run the test case. To make sure setup is correct, check
the ```layout.md``` file located in __system/mesh/layout/layout.md__. The file should look like the following:
```
## BLOCK LAYOUT FILE
## ==========================
## NUMBER OF PROCESSES
6
## NUMBER OF ENTRIES PER PROCESS
9
## PROCESS_NO GRID BC_FILE IMIN IMAX JMIN JMAX KMIN KMAX
## ===================================
## PROCESS 0
00  grid_00.txt  bc_00.md  -004  0001  0005  -008  -006  -006
## PROCESS 1
01  grid_01.txt  bc_01.md  0000  0002  -005  -008  -006  -006
## PROCESS 2
02  grid_02.txt  bc_02.md  0001  0003  -005  -008  -006  -006
## PROCESS 3
03  grid_03.txt  bc_03.md  0002  0004  -005  -008  -006  -006
## PROCESS 4
04  grid_04.txt  bc_04.md  0003  0005  -005  -008  -006  -006
## PROCESS 5
05  grid_05.txt  bc_05.md  0004  -004  0000  -008  -006  -006
```

Finally, to run the simulation use the following command:
```
$nohup bash run.sh &
```
`nohup` helps in avoiding any output on the screen, __&__ execute the last command in the background, allowing you to keep using the terminal.

